"""
RealTimeROG.py

    Uses a webcam to measure and plot the radius of gyration of a beaded chain
        in real time.
    Windows-Only: requires VideoCapture - http://videocapture.sourceforge.net
    Also requires:
        Python Imaging Library (PIL): http://www.pythonware.com/products/pil/
    
    Copyright (c) 2011 University of Toronto
    Original Version :  12 August 2011 by Donald J. Woodbury 
    Contact : David Bailey <dbailey@physics.utoronto.ca>
                (www.physics.utoronto.ca/~dbailey)
    License: Released under the MIT License; the full terms are appended
                to the end of this file, and are also available
                at www.opensource.org/licenses/mit-license.php
"""

from Tkinter import *
from RTRogTools import *
from tkSimpleDialog import askstring
from tkFileDialog import asksaveasfilename, askopenfilename
from tkMessageBox import askokcancel
from visual.graph import gcurve, color
from time import clock, sleep, strftime
from PIL import Image, ImageTk
import os
import pylab

from VideoCapture import Device

class Run_webcam:
    '''Launches a TkInter Window that will display the current stream from a
    webcam linked to the current computer.'''
    def __init__(self):

        #Initializing the webcam

        self.cam = Device()

        #Initializing the GUI
        
        self.root = Tk()
        self.root.title('Video Capture')

        #Initialize parameters
        
        self.cam.setResolution(1600, 1200) #Image Resolution
        self.xsize, self.ysize = (640, 480) #displayed frame size

        self.t = clock()
        self.run_camera = 0
        self.plot_ROG = 0
        self.save_ROG = 0
        self.resizing = 0
        self.record_images = 0
        self.bbox = []
        self.pix_to_cm = [1,1]

        #Building the Tkinter window

        self.Create_window()

        #Building the Tkinter Menus

        self.menu = Menu(self.root)
        self.root.config(menu=self.menu)

        self.File_menu()
        self.Camera_menu()
        self.Help_menu()

        #Building the rest of the window

        self.Sliders()
        self.Bindings()

        self.root.mainloop()

    def Create_window(self):
        '''Inserts the first image into the Tkinter dialog'''

        self.image = self.cam.getImage().convert('L')
        
        self.scaled_image = Resize_image(self.image, self.xsize, self.ysize)

        self.tk_image = ImageTk.PhotoImage(self.scaled_image)

        self.frame = Frame(self.root, width = self.xsize, height = self.ysize)
        self.frame.grid(row = 0, sticky = 'n', columnspan = 5)

        self.canvas = Canvas(self.frame, width = self.xsize, height = self.ysize)
        self.canvas.pack()

        self.im_id = self.canvas.create_image(0,0,image=self.tk_image,anchor="nw")

    #_____________________File Menu Functions_____________________#

    def File_menu(self):
        '''Builds the File menu'''
        self.filemenu = Menu(self.menu)
        
        self.menu.add_cascade(label="File", menu=self.filemenu)
        self.find_chain = IntVar(0)
        self.filemenu.add_checkbutton(label="Find Chain", \
                                      command=self.Find_chain,
                                      variable = self.find_chain)
        self.filemenu.add_command(label="Start Plotting", \
                               command=self.Start_Plot)
        self.filemenu.add_separator()
        self.filemenu.add_checkbutton(label="Save ROG to File",\
                                      command=self.Save_ROG)
        self.filemenu.add_command(label="Take Snapshot",\
                                      command=self.Take_snapshot)
        self.filemenu.add_checkbutton(label="Record Images",\
                                      command=self.Record_images)
        self.filemenu.add_command(label="Plot from File",\
                                      command=self.Pylab_plot)
        self.filemenu.add_separator()
        self.filemenu.add_command(label="Exit", command=self.End_loop)

    def Find_chain(self):
        '''If Chain detection is off, this function turns it on. Otherwise
        it turns it off and if the track is being saved, it closes the file.'''
        if not self.find_chain.get():
            self.save_ROG = 0
            try:
                self.ROG_file.close()
            except AttributeError:
                pass
        else:
            if not self.run_calibration.get():
                if not askokcancel("Find Spot", 'Warning: without first\n'+\
                                   'defining the plate boundaries\n'+\
                                   'the program may run slowly', \
                                   icon = 'warning'):
                    self.find_chain.set(0)
            
    def Start_Plot(self):
        '''Activates the real-time vPython plot, or pauses it'''
        if self.plot_ROG == 0:
            try:
                self.start_t
            except AttributeError:
                self.start_t = clock()
            self.f1 = gcurve(color=color.cyan)
            self.plot_ROG = 1
            self.filemenu.entryconfig("Start Plotting", label = "Stop Plotting")
        else:
            self.plot_ROG = 0
            self.filemenu.entryconfig("Stop Plotting", label = 'Start Plotting')

    def Save_ROG(self):
        '''Prompts the user to choose a filename underwhich to save the data
        and then sets self.save_ROG to true so that the track location will be
        saved with every frame.'''
        if self.save_ROG:
            self.save_ROG = 0
            self.ROG_file.close()
        else:
            filename = asksaveasfilename(parent=self.root, filetypes = [('Text File','*.txt')],\
                                         title="Save the file as...")
            if len(filename) > 0:
                try:
                    self.start_t
                except AttributeError:
                    self.start_t = clock()
                
                self.ROG_file = open(filename, 'w')
                self.ROG_file.write('Time(s)\tROG(pix)\n\n')
                self.save_ROG = 1
                
    def Take_snapshot(self):
        '''creates a new directory called 'Snapshots' if it does not already
        exist, and then saves the current frame, indicating the date and time
        in the format yyyymmdd_HrMinSec. If either vPython plotting has been
        initialized, or if the ROG is being saved to a file, then "_t" is
        added to the name where t is the time since the first of those two was
        started.'''
        directory = os.getcwd() + '\\ROG_Snapshots'
        if not os.path.exists(directory):
            os.makedirs(directory)
        try:
            self.image.save(directory +'\ROG_Snapshot_' + \
                            strftime("%Y%m%d_%H%M%S") + "_%.2f.png" \
                            %(clock()-self.start_t))
        except AttributeError:
            self.image.save(directory +'\ROG_Snapshot_' + \
                            strftime("%Y%m%d_%H%M%S") + '.png')
        try:
            print 'Snapshot at: %.2f' % (clock()-self.start_t)
        except AttributeError:
            print 'Snapshot at: %.2f' % clock()
    def Record_images(self):
        '''Takes a snapshot at every user defined interval.'''
        if self.record_images:
            self.record_images = 0
        else:
            self.save_frame_rate = float(askstring('Enter frame rate', \
                                            'Enter the time between images (s):'))
            self.Take_snapshot()
            self.last_snapshot_t = clock()
            self.record_images = 1

    def Pylab_plot(self):
        '''Launches a Toplevel window that prompts the user for parameters to
        display a pylab plot of the data saved to a file.'''
        Plot_ROG(self.root)

    #_____________________Camera Menu Functions_____________________#

    def Camera_menu(self):
        '''Builds the camera menu'''
        self.cameramenu = Menu(self.menu)
        
        self.menu.add_cascade(label="Camera", menu=self.cameramenu)
        self.cameramenu.add_command(label="Start Camera", \
                               command=self.Start_Webcam)
        self.cameramenu.add_separator()
        self.cameramenu.add_command(label="Webcam Controls", \
                               command=self.Control_webcam)
        self.cameramenu.add_command(label="Adjust Resolution", \
                               command=self.Change_resolution)
        self.run_calibration = IntVar()
        self.cameramenu.add_checkbutton(label="Calibrate",\
                                        command = self.Calibrate,
                                        variable = self.run_calibration)

    def Calibrate(self):
        '''Prompts the user to select the bounding sides of the plate, and specify the
        diameter of the plate. Once that is initialized, the program only searches the
        plate for the chain and displays all ROG data in whatever unit was used to
        measure the diameter.'''
        
        if self.run_calibration.get() == 1:
            calibration_data = Calibration(self.root, self.image)
            self.bbox = calibration_data.corners
            self.pix_to_cm = calibration_data.pix_to_cm
        else:
            self.bbox = []
            self.pix_to_cm = [1,1]

    def Control_webcam(self):
        '''Launches the control panel for the exposure, focus, zoom, etc.'''
        self.cam.displayCaptureFilterProperties()
    def Change_resolution(self):
        '''Launches the control panel for the Resolution.'''
        size = self.image.size
        self.cam.displayCapturePinProperties()

        if self.cam.getImage().size != size:
            self.find_chain.set(0)
            self.Find_chain()
            self.run_calibration.set(0)
            self.Calibrate()

    #_____________________Help Menu________________________#

    def Help_menu(self):
        '''Builds the Help menu'''
        helpmenu = Menu(self.menu)
        self.menu.add_cascade(label="Help", menu=helpmenu)

    #_____________________Sliders________________________#

    def Sliders(self):
        '''Initializes all sliders contained in the Tkinter window.'''

        self.threshold = IntVar()
        self.threshold.set(200)
        self.threshold_slider = Scale(self.root, variable = self.threshold, \
                                 label="Threshold:",length=self.xsize, \
                                 orient=HORIZONTAL, to=255)
        self.threshold_slider.grid(row = 1, sticky = 'w', columnspan = 5)

    #_____________________Bindings_____________________#

    def Bindings(self):
        self.root.protocol("WM_DELETE_WINDOW", self.End_loop)
        self.root.bind("<Configure>", self.Change_window_size)

    def Change_window_size(self, event):
        '''Resizes the image and slider when the user resizes window.'''
        if clock() > 1 and not self.resizing:
            self.resizing = 1
            self.window_size = (int(self.root.winfo_reqwidth()),\
                                int(self.root.winfo_reqheight()))
            self.xsize = self.xsize + (event.width - self.window_size[0])
            self.ysize = self.ysize + (event.height - self.window_size[1])
        
            self.canvas.config(width = self.xsize, height = self.ysize)
            self.threshold_slider.config(length=self.xsize)
            self.root.update()
            self.resizing = 0
    
    def End_loop(self):
        '''Closes the window and, if the ROG is being saved to a file,
        closes the ROG file'''
        if askokcancel("Quit", "Do you really wish to quit?"):
            self.run_camera = 0
            self.root.destroy()
            try:
                self.ROG_file.close()
            except AttributeError:
                pass

    #___________________Run Webcam and Analysis__________________#

    def Start_Webcam(self):
        '''Runs the webcam continually, while performing the tasks defined by
        the parameters set above.'''
        
        if self.run_camera == 0:
            self.run_camera = 1
            self.cameramenu.entryconfig("Start Camera", label = 'Stop Camera')
        else:
            self.run_camera = 0
            self.cameramenu.entryconfig("Stop Camera", label = 'Start Camera')
            self.save_ROG = 0
            try:
                self.ROG_file.close()
            except AttributeError:
                pass

        while self.run_camera:
            self.t = clock()
            self.image = self.cam.getImage().convert('L')
            
            if self.find_chain.get():
                self.Analysis_Functions() #Video Analysis functions
                
            if self.record_images and \
               (clock() - self.last_snapshot_t) > self.save_frame_rate:
                self.Take_snapshot()
                self.last_snapshot_t = clock()
                
            self.scaled_image = Resize_image(self.image, self.xsize, self.ysize)
            
            self.canvas.delete(self.im_id)
            self.tk_image = ImageTk.PhotoImage(self.scaled_image)
            self.im_id = self.canvas.create_image(0,0,image=self.tk_image,anchor="nw")
            self.root.update()
        
    def Analysis_Functions(self):
        '''Finds the Radius of gyration and displays, saves or plots it depending on the
        parameters set above.'''

        self.points = points_above_threshold(self.image, self.threshold.get(), self.bbox)

        if self.points != []:
            self.ROG = radius_of_gyration(self.points, self.pix_to_cm)
            self.image = draw_points(self.points, self.image)
            if self.plot_ROG:
                self.f1.plot(pos=(self.t-self.start_t,self.ROG))
            if self.save_ROG:
                self.ROG_file.write('%.2f\t%.2f\n' % (self.t-self.start_t,\
                                                      self.ROG))
                print '%.2f\t%.2f' % (self.t-self.start_t, self.ROG)       
        else:
            print 'Cannot find chain at time %.2f' % self.t

if __name__ == '__main__':
    Run_webcam()
